//---------------------------------------------------------------------------
//
//  Program:     UnitOpenDialog for MidiAndMusicXmlPlayer
//
//  Project:     MidiAndMusicXmlPlayer.exe and for MidiAndMusicXmlPlayer.app
//
//  Purpose:     To select a Midi or Music Xml file by showing only files
//               in the Music and Download folder and only xml files
//               that are Music Xml files
//
//  Compilation: Compile with Delphi 4 (MidiAndMusicXmlPlayer.dpr),
//               Delphi XE (MidiAndMusicXmlPlayer.dproj)
//               or Lazarus (MidiAndMusicXmlPlayer.lpi)
//
//  Description: A tree view is build to show and select files
//               All Xml files are opened and must contain the word
//               MusicXML in the second line of the file
//---------------------------------------------------------------------------

unit UnitOpenDialog;

{$mode delphi}
{$H+}

interface

uses
{$ifdef Darwin}
  CWString,
{$endif}
  Classes, SysUtils, FileUtil, Forms, Controls, Graphics, Dialogs, ComCtrls,
  ExtCtrls, StdCtrls, Buttons,
  UnitMidiDefinitions,UnitMessage;

// Some space around boxes and buttons
const GLUE = 5;

const
  SAnyMask = {$ifdef Darwin}'*'{$else}'*.*'{$endif};
  PathDelim = {$ifdef Darwin}'/'{$else}'\'{$endif};
  faSymLink = $00000040;

const
  // Select the icons used from the Icon bitmap
  // Note: 0 = yellow folder
  //       2 = yellow Win8 like
  IconFolder = 3;  // Blue folder
  IconFile = 1;    // The Midi/Music Xml icon

type

  { TFormOpenDialog }

  TFormOpenDialog = class(TForm)
    BevelTreeView: TBevel;
    EditFileName: TEdit;
    ImageButtonOkNew: TImage;
    ImageFolders: TImage;
    Timer: TTimer;
    TreeView: TTreeView;
    procedure ImageButtonOkNewClick(Sender: TObject);
    procedure OnMouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer);
    procedure OnEdited(Sender: TObject; Node: TTreeNode; var S: string);
    procedure OnActivate(Sender: TObject);
    procedure OnClose(Sender: TObject; var CloseAction: TCloseAction);
    procedure OnCreate(Sender: TObject);
    procedure OnDblClick(Sender: TObject);
    procedure OnResize(Sender: TObject);
    procedure TimerTimer(Sender: TObject);
    procedure OnSelectionChanged(Sender: TObject);
  private
    { private declarations }
    SelectedFileName: string;
    SetVisible: boolean;
    procedure FindFiles(Dir: string; RootNode: TTreeNode);
    function DirNotEmpty(Dir: string; RootNode: TTreeNode): boolean;
    function AddNode(Node: TTreeNode; Name: string): TTreeNode;
    function AddSubNode(Node: TTreeNode; Name: string;
                        IconNr: integer; Description: string): TTreeNode;
    function ShortPath(Name: string): string;
    procedure SavePosition;
  public
    { public declarations }
  end;

var
  FormOpenDialog: TFormOpenDialog;

implementation

{$R *.lfm}

{ TFormOpenDialog }

//---------------------------------------------------------------------------
//
//     Function:   AddNode
//
//     Purpose:    To add a node to the tree view. Show a folder icon.
//
//     Parameters: Node = the tree view node
//                 Name = text shown (music or download)
//
//     Returns:    void
//
//     Notes:      <none>
//
//---------------------------------------------------------------------------

function TFormOpenDialog.AddNode(Node: TTreeNode; Name: string): TTreeNode;
var
  NewNode: TTreeNode;

  begin
  NewNode:=TreeView.Items.Add(Node,Name);
  NewNode.ImageIndex:=IconFolder;
  NewNode.SelectedIndex:=IconFolder;
  NewNode.Data:=nil;
  AddNode:=NewNode;
  end;

//---------------------------------------------------------------------------
//
//     Function:   AddSubNode
//
//     Purpose:    To add a subnode to the tree view. Show a folder or
//                 file icon.
//
//     Parameters: Node = the tree view node
//                 Name = text shown (music or download)
//                 IconNr = The icon according to the icon file.
//                 (May be different on PC and MAC. MAC is p.t. blue)
//                 See definitions IconFolder and IconFile
//
//     Returns:    void
//
//     Notes:      <none>
//
//---------------------------------------------------------------------------
function TFormOpenDialog.AddSubNode(Node: TTreeNode; Name: string;
  IconNr: integer; Description: string): TTreeNode;

var
  RootNode: TTreeNode;
  s: string;
  PDescription: ^string;

  begin
  RootNode:=Node;
  Node:=TreeView.Items.AddChild(RootNode,Name);
  Node.ImageIndex:=IconNr;
  Node.SelectedIndex:=IconNr;
  New(PDescription);                 //// OBS: Add freeing the memory !!!!
  PDescription^:=Description;
  Node.Data:=PDescription;
  RootNode.Expand(true);
  AddSubNode:=Node;
  end;

//---------------------------------------------------------------------------
//
//     Function:   OnCreate
//
//     Purpose:    To handle the standard windows event OnCreate
//                 If a file Default.txt exists then this file contains
//                 a position and size of this window. After this initialise
//                 the tree view of the directories for Music and Download
//
//     Parameters: Sender = Windows standard
//
//     Returns:    void
//
//     Notes:      <none>
//
//---------------------------------------------------------------------------

procedure TFormOpenDialog.OnCreate(Sender: TObject);
var
  ImageList: TImageList;
  Home: string;
  WindowPosition: TextFile;
  L,T,W,H: integer;           // Local variables: Left, Top, Width, Height

  NodeMusicTop: TTreeNode;
  NodeDownloadTop: TTreeNode;

  begin
  L:=Left;
  T:=Top;
  W:=Width;
  H:=Height;
  SelectedFileName:='';
  EditFileName.Text:='';
  Home:=GetEnvironmentVariable('HOME');
  if FileExists(Home+'/MidiOptions/Default.txt') then
    begin
    AssignFile(WindowPosition,Home+'/MidiOptions/Default.txt');
    reset(WindowPosition);
    if not eof(WindowPosition) then readln(WindowPosition,L);
    if not eof(WindowPosition) then readln(WindowPosition,T);
    if not eof(WindowPosition) then readln(WindowPosition,W);
    if not eof(WindowPosition) then readln(WindowPosition,H);
    CloseFile(WindowPosition);
    end;
  Left:=L;
  Top:=T;
  Width:=W;
  Height:=H;
  ImageList:=TImageList.Create(nil);
  ImageList.Clear;
  ImageList.Add(ImageFolders.Picture.Bitmap,ImageFolders.Picture.Bitmap);
  TreeView.Images:=ImageList;
  TreeView.Items.Clear;
  NodeMusicTop:=AddNode(nil,'Music');
  NodeDownloadTop:=AddNode(nil,'Download');
  FindFiles(Home+'/Music/',NodeMusicTop);
  FindFiles(Home+'/Downloads/',NodeDownloadTop);
  SelectedFileName:='';
  TreeView.FullExpand;
  end;

//---------------------------------------------------------------------------
//
//     Function:   OnDblClick
//
//     Purpose:    To handle the standard windows event OnDblClick
//                 The file is opened directly
//
//     Parameters: Sender = Windows standard
//
//     Returns:    void
//
//     Notes:      <none>
//
//---------------------------------------------------------------------------

procedure TFormOpenDialog.OnDblClick(Sender: TObject);

var
  Node: TTreeNode;
  PDescription: ^string;
  CloseAction: TCloseAction;

  begin
  Errors:=0;
  Node:=TreeView.Selected;
  if Node<>nil then
    begin
    PDescription:=Node.Data;
    if (PDescription<>nil) and (PDescription^<>'') and
       ((LowerCase(ExtractFileExt(PDescription^))='.mid') or
          (LowerCase(ExtractFileExt(PDescription^))='.xml')) then
      begin
      SelectedFileName:=PDescription^;
      EditFileName.Text:=ShortPath(PDescription^);
      ImageButtonOkNewClick(Sender);
      end
    else
      begin
      SelectedFileName:='';
      EditFileName.Text:='';
      end;
    end;
  SavePosition;
  end;

//---------------------------------------------------------------------------
//
//     Function:   OnResize
//
//     Purpose:    To handle the standard windows event OnResize
//                 This window can be of any size and the elements
//                 are resized accordingly
//
//     Parameters: Sender = Windows standard
//
//     Returns:    void
//
//     Notes:      <none>
//
//---------------------------------------------------------------------------

procedure TFormOpenDialog.OnResize(Sender: TObject);

    begin
    VertScrollBar.Visible:=false;

    ImageButtonOkNew.Top:=ClientHeight-(3*EditFileName.Height) div 2;
    ImageButtonOkNew.Width:=ClientWidth div 10;
    ImageButtonOkNew.Height:=EditFileName.Height;
    ImageButtonOkNew.Left:=ClientWidth-ImageButtonOkNew.Width-6*GLUE;

    EditFileName.Top:=ClientHeight-(3*EditFileName.Height) div 2;
    EditFileName.Left:=2*GLUE;
    EditFileName.Width:=(8*ClientWidth) div 10;
    TreeView.Top:=2*GLUE;
    TreeView.Height:=ClientHeight-2*ImageButtonOkNew.Height-3*GLUE;
    TreeView.Left:=2*GLUE;
    TreeView.Width:=ClientWidth-4*GLUE;

    BevelTreeView.Top:=Glue;
    BevelTreeView.Left:=Glue;
    BevelTreeView.Width:=ClientWidth-2*Glue;
    BevelTreeView.Height:=TreeView.Height+2*Glue;
{
    BevelFileName.Top:=ClientHeight-2*EditFileName.Height+GLUE;
    BevelFileName.Left:=Glue;
    BevelFileName.Width:=ClientWidth-2*Glue;
    BevelFileName.Height:=2*EditFileName.Height-2*GLUE;
}
    end;


procedure TFormOpenDialog.OnActivate(Sender: TObject);

  begin
  //// P.t. used in the PC-version but...
  end;

//---------------------------------------------------------------------------
//
//     Function:   OnClose
//
//     Purpose:    To handle the standard windows event OnClose
//                 The position and size of the window is saved
//
//     Parameters: Sender = Windows standard
//
//     Returns:    void
//
//     Notes:      <none>
//
//---------------------------------------------------------------------------

procedure TFormOpenDialog.OnClose(Sender: TObject; var CloseAction: TCloseAction);

  begin
  SavePosition;
  end;

//---------------------------------------------------------------------------
//
//     Function:   SavePosition
//
//     Purpose:    The save the position and size of the window
//                 in order to show it at same place next time
//
//     Parameters: <none>
//
//     Returns:    void
//
//     Notes:      <none>
//
//---------------------------------------------------------------------------

procedure TFormOpenDialog.SavePosition;
  var
    WindowPosition: TextFile;
    Home: string;

    begin
    Home:=GetEnvironmentVariable('HOME');
    if not DirectoryExists(Home+'/MidiOptions') then
      MkDir(Home+'/MidiOptions');
    AssignFile(WindowPosition,Home+'/MidiOptions/Default.txt');
    rewrite(WindowPosition);
    writeln(WindowPosition,Left);
    writeln(WindowPosition,Top);
    writeln(WindowPosition,Width);
    writeln(WindowPosition,Height);
    CloseFile(WindowPosition);
    end;

//---------------------------------------------------------------------------
//
//     Function:   ButtonOkClick
//
//     Purpose:    To handle the standard windows event ButtonOkClick
//                 By mouse click the the is selected and show in the edit
//                 field below the file tree view. The file is opened directly
//                 The system state is changed to be ready to start playing
//
//     Parameters: Sender = Windows standard
//
//     Returns:    void
//
//     Notes:      <none>
//
//---------------------------------------------------------------------------
{
procedure TFormOpenDialog.ImageButtonOkNewClick(Sender: TObject);

    begin
    Errors:=0;
    if FileExistsUtf8(SelectedFileName) then
      begin
      FileName:=SelectedFileName;
      FormMessage.Create;
      SystemState:=MidiStarting;
      end
    else
      begin
      SystemState:=MidiNoFile;
      end;
    Visible:=false;
    SavePosition;
    end;
 }

//---------------------------------------------------------------------------
//
//     Function:   ImageButtonOkClick
//
//     Purpose:    To handle the standard windows event ButtonOkClick
//                 By mouse click the the is selected and show in the edit
//                 field below the file tree view. The file is opened directly
//                 The system state is changed to be ready to start playing
//
//     Parameters: Sender = Windows standard
//
//     Returns:    void
//
//     Notes:      <none>
//
//---------------------------------------------------------------------------

procedure TFormOpenDialog.ImageButtonOkNewClick(Sender: TObject);

  begin
  Errors:=0;
  if FileExistsUtf8(SelectedFileName) then
    begin
    FileName:=SelectedFileName;
    FormMessage.Create;
    SystemState:=MidiStarting;
    end
  else
    begin
    SystemState:=MidiNoFile;
    end;
  Visible:=false;
  SavePosition;
  end;

//---------------------------------------------------------------------------
//
//     Function:   OnMouseMove
//
//     Purpose:    To handle the standard windows event OnMouseMove
//                 NOT USED. PREPARE DRAG AND DROP
//
//     Parameters: Sender = Windows standard
//                 shift  = state of CTRL and ALT and SHIFT buttons
//                 X      = x-position of mouse pressed
//                 Y      = y-position of mouse pressed
//
//     Returns:    void
//
//     Notes:      <none>
//
//---------------------------------------------------------------------------

procedure TFormOpenDialog.OnMouseMove(Sender: TObject; Shift: TShiftState; X,
  Y: Integer);

var DragMode: TDragMode;
    DragKind: TDragKind;
    FileName: string;
    ////FileObject: TFileObject;

  begin
  ////DataObject:=GetFileListDataObject(Directory, SelFileList);
  ////DoDragDrop(DataObject, Self, DROPEFFECT_COPY, Effect);
  ////CopyFilesToClipboard();

  TreeView.DragMode:=dmManual;
  TreeView.DragCursor:=crDrag;
  BeginDrag(true);
  TreeView.DragCursor:=crDrag;
  TreeView.DragKind:=dkDrag;
  ////TreeView.DragDrop(FileName,1,1);
  end;

//---------------------------------------------------------------------------
//
//     Function:   OnEdited
//
//     Purpose:    To handle the standard windows event OnEdited
//                 A name in the tree view may be edited. Change
//                 the file name to the new one
//
//     Parameters: Sender = Windows standard
//
//     Returns:    void
//
//     Notes:      <none>
//
//---------------------------------------------------------------------------

procedure TFormOpenDialog.OnEdited(Sender: TObject; Node: TTreeNode;var S: string);

  var
    OldFileName: string;         //// description !!!!
    NewFileName: string;
    OldFilePath: string;
    NewFilePath: string;
    PDescription: ^string;
    p: integer; // Position of delimiter '/'
    i: integer; // Loop variable  for positions

    begin
    PDescription:=Node.Data;
    OldFilePath:=PDescription^;
    OldFileName:=ExtractFileName(OldFilePath);
    // Find the path part
    p:=0;
    for i:=1 to Length(OldFilePath) do if OldFilePath[i]='/' then p:=i;
    NewFilePath:=copy(OldFilePath,1,p);
    NewFileName:=S;
    NewFilePath:=NewFilePath+NewFileName;
    RenameFile(OldFilePath,NewFilePath);
    PDescription^:=NewFilePath;
    SelectedFileName:=PDescription^;
    EditFileName.Text:=ShortPath(PDescription^);
    end;

//---------------------------------------------------------------------------
//
//     Function:   TimerTimer
//
//     Purpose:    To handle the standard windows event TimerTimer
//                 When a tree item is selected, there seem to be an
//                 updating problem - sometimes a wrong name appears
//                 (the previous selected item). Therefore the timer
//                 assures that the name will be right at next timer
//                 click.
//
//     Parameters: Sender = Windows standard
//
//     Returns:    void
//
//     Notes:      <none>
//
//---------------------------------------------------------------------------

procedure TFormOpenDialog.TimerTimer(Sender: TObject);
var
  Node: TTreeNode;
  PDescription: ^string;

  begin
  Node:=TreeView.Selected;
  if Node<>nil then
    begin
    PDescription:=Node.Data;
    if (PDescription<>nil) and (PDescription^<>'') and
       ((LowerCase(ExtractFileExt(PDescription^))='.mid') or
          (LowerCase(ExtractFileExt(PDescription^))='.xml')) then
      begin
      SelectedFileName:=PDescription^;
      EditFileName.Text:=ShortPath(PDescription^);
      end;
    end;
  end;

procedure TFormOpenDialog.OnSelectionChanged(Sender: TObject);

  begin
////  ImageButtonOkNew.Refresh; //// ????
  end;



function StrAppend(var s1: string; var s2: string): string;

var s: string;
    i: integer;

  begin
  s:='';
  for i:=1 to Length(s1) do
    s:=s+s1[i];
  for i:=1 to Length(s2) do
    s:=s+s2[i];
  StrAppend:=s;
  end;

//---------------------------------------------------------------------------
//
//     Function:   FindFiles
//
//     Purpose:    To traverse the whole file structure and add the
//                 elements to the tree view.
//
//     Parameters: Dir = current directory
//                 RootNode = current node in the tree view
//
//     Returns:    void
//
//     Notes:      <none>
//
//---------------------------------------------------------------------------

procedure TFormOpenDialog.FindFiles(Dir: string; RootNode: TTreeNode);

var
  Info: TSearchRec;
  FileName: string;
  DirName: string;
  InfoName: string;
  FileNameWithPath: AnsiString;
  Node: TTreeNode;

  MusicXmlFile: textfile;
  MidiFile: textfile;
  s: string;
  c1: char;
  c2: char;
  c3: char;
  c4: char;

  i: integer;

  begin
{
  writeln('DefaultSystemCodePage:',DefaultSystemCodePage,
          ' DefaultFileSystemCodePage=',DefaultFileSystemCodePage,
          ' DefaultRTLFileSystemCodePage=',DefaultRTLFileSystemCodePage);
  writeln('LC_ALL=',GetEnvironmentVariable('LC_ALL'),' LC_MESSAGES=',
          GetEnvironmentVariable('LC_MESSAGES'),
          ' LANG=',GetEnvironmentVariable('LANG'));
}
  // Find files in the current directory
  if FindFirstUTF8(Dir+FileName,faAnyFile and not faDirectory,Info) { *Converted from FindFirst* } = 0 then
  try
    repeat
      // Work with found file info...
    until FindNextUTF8(Info) { *Converted from FindNext* }<>0;
  finally
    FindCloseUTF8(Info); { *Converted from FindClose* }
  end;
  // Find subdirectories and recurse
  if FindFirstUTF8(Dir+SAnyMask,faDirectory,Info) { *Converted from FindFirst* }=0 then
    try
      repeat
      if (Info.Attr and faDirectory<>0) and
         (Info.Name<>'.') and (Info.Name<>'..') and
         (Info.Attr and faSymLink=0) then
        begin // Dir - add name to directory path
        if DirNotEmpty(Dir+Info.Name+PathDelim,Node) then
        Node:=AddSubNode(RootNode,Info.Name,IconFolder,Dir+Info.Name);
        FindFiles(Dir+Info.Name+PathDelim,Node);
        end
      else
        begin // File - only midi and music xml.
        if ((Info.Name<>'.') and (Info.Name<>'..')) and
           (LowerCase(ExtractFileExt(Info.Name))='.mid') or
           (LowerCase(ExtractFileExt(Info.Name))='.xml') then
          begin
          if LowerCase(ExtractFileExt(Info.Name))='.xml' then
            begin
            {$i-}
            InfoName:=Info.Name;
            DirName:=Dir;
////            FileName:=DirName+InfoName;
            FileName:=StrAppend(DirName,InfoName);

////            FileNameWithPath:=Utf8ToAnsi(FileName);
            AssignFile(MusicXmlFile,Dir+Info.Name);
////            ChDir(Dir);
////            AssignFile(MusicXmlFile,Info.Name);
////            AssignFile(MusicXmlFile,FileNameWithPath);
            if IORESULT=0 then
            reset(MusicXmlFile);
            {$i+}
            if IORESULT=0 then
              begin
              // A music xml file has the word MusicXML in second line
              if not eof(MusicXmlFile) then readln(MusicXmlFile,s);
              if not eof(MusicXmlFile) then readln(MusicXmlFile,s);
              if Pos('MusicXML',s)>0 then
              Node:=AddSubNode(RootNode,Info.Name,IconFile,Dir+Info.Name);
              CloseFile(MusicXmlFile);
              end;
            end
          else
            begin // Midi File - Extension = ".mid"
            if FileExists(Dir+Info.Name) then
              begin
              AssignFile(MidiFile,Dir+Info.Name);
              if IORESULT=0 then
              reset(MidiFile);
              {$i+}
              if IORESULT=0 then
                begin
                // A midi file has the characters "MThd" in the start
                if not eof(MidiFile) then read(MidiFile,c1);
                if not eof(MidiFile) then read(MidiFile,c2);
                if not eof(MidiFile) then read(MidiFile,c3);
                if not eof(MidiFile) then read(MidiFile,c4);
                if (c1='M') and (c2='T') and (c3='h') and (c4='d') then
                Node:=AddSubNode(RootNode,Info.Name,IconFile,Dir+Info.Name);
                CloseFile(MidiFile);
                end;
              end;
            end;
          end;
        end;
      until FindNextUTF8(Info) { *Converted from FindNext* }<>0;
    finally
    FindCloseUTF8(Info); { *Converted from FindClose* }
    end;
  end;

//---------------------------------------------------------------------------
//
//     Function:   DirNotEmpty
//
//     Purpose:    To traverse the whole file structure and test
//                 if directory is empty (i.e. contains no Midi nor Music Xml)
//
//     Parameters: Dir = current directory
//                 RootNode = current node in the tree view
//
//     Returns:    void
//
//     Notes:      <none>
//
//---------------------------------------------------------------------------

function TFormOpenDialog.DirNotEmpty(Dir: string; RootNode: TTreeNode): boolean;

var
  Info: TSearchRec;
  FileName: string;
  Node: TTreeNode;
  i: integer;
  NotEmpty: boolean;
  MusicXmlFile: textfile;
  s: string;

  begin
  // Find files in the current directory...
  if FindFirstUTF8(Dir+FileName,faAnyFile and not faDirectory,Info) { *Converted from FindFirst* } = 0 then
  try
    repeat
      // Work with found file info...
    until FindNextUTF8(Info) { *Converted from FindNext* }<>0;
  finally
    FindCloseUTF8(Info); { *Converted from FindClose* }
  end;
  // Find subdirectories and recurse
  if FindFirstUTF8(Dir+SAnyMask,faDirectory,Info) { *Converted from FindFirst* }=0 then
    try
      NotEmpty:=false;
      repeat
      if (Info.Attr and faDirectory<>0) and
         (Info.Name<>'.') and (Info.Name<>'..') and
         (Info.Attr and faSymLink=0) then
        begin // Dir
        NotEmpty:=NotEmpty or DirNotEmpty(Dir+Info.Name+PathDelim,Node);
        end
      else
        begin // File
{
        FileName:=Dir+'xyzæøå.test';
        AssignFile(MusicXmlFile,FileName);
        rewrite(MusicXmlFile);
        CloseFile(MusicXmlFile);
}
        if ((Info.Name<>'.') and (Info.Name<>'..')) and
           (LowerCase(ExtractFileExt(Info.Name))='.mid') or
           (LowerCase(ExtractFileExt(Info.Name))='.xml') then
          begin
          if LowerCase(ExtractFileExt(Info.Name))='.xml' then
            begin
            {$i-}
            AssignFile(MusicXmlFile,PChar(Dir+Info.Name));
            if IORESULT=0 then
            reset(MusicXmlFile);
            {$i+}
            if IORESULT=0 then
              begin
              if not eof(MusicXmlFile) then readln(MusicXmlFile,s);
              if not eof(MusicXmlFile) then readln(MusicXmlFile,s);
              if Pos('MusicXML',s)>0 then NotEmpty:=true;
              CloseFile(MusicXmlFile);
              end;
            end
          else
            begin
            if FileExists(Dir+Info.Name) then NotEmpty:=true;
            end;
          end;
        end;
      until FindNextUTF8(Info) { *Converted from FindNext* }<>0;
    finally
    FindCloseUTF8(Info); { *Converted from FindClose* }
    end;
    DirNotEmpty:=NotEmpty;
  end;

//---------------------------------------------------------------------------
//
//     Function:   ShortPath
//
//     Purpose:    In order to avoid the very long and strange
//                 names in the Apple's sandboxing environment
//                 only show 'Music' and 'Download' and the rest
//
//     Parameters: Name = the real file name with path
//
//     Returns:    A more readable name.
//
//     Notes:      <none>
//
//---------------------------------------------------------------------------

function TFormOpenDialog.ShortPath(Name: string): string;

var
  Home: string;
  p: integer;

  begin
  Home:=GetEnvironmentVariable('HOME');
  p:=Pos(Home,Name);
  // Pass the HOME and '/'
  if p=1 then ShortPath:=copy(Name,Length(Home)+2,Length(Name)-Length(Home));
  end;


end.



//---------------------------------------------------------------------------
//
//  Program:     UnitOpenDialog for MidiAndMusicXmlPlayer
//
//  Project:     MidiAndMusicXmlPlayer.exe and for MidiAndMusicXmlPlayer.app
//
//  Purpose:     To select a Midi or Music Xml file by showing only files
//               in the Music and Download folder and only xml files
//               that are Music Xml files
//
//  Compilation: Compile with Delphi 4 (MidiAndMusicXmlPlayer.dpr),
//               Delphi XE (MidiAndMusicXmlPlayer.dproj)
//               or Lazarus (MidiAndMusicXmlPlayer.lpi)
//
//  Description: A tree view is build to show and select files
//               All Xml files are opened and must contain the word
//               MusicXML in the second line of the file
//---------------------------------------------------------------------------

unit UnitOpenDialog;

{$mode delphi}
{$H+}

interface

uses
  CWString,
  Classes, SysUtils, FileUtil, Forms, Controls, Graphics, Dialogs, ComCtrls,
  ExtCtrls, StdCtrls,
  UnitMidiDefinitions,UnitMessage;

// Some space around boxes and buttons
const GLUE = 5;

const
  SAnyMask = {$ifdef Darwin}'*'{$else}'*.*'{$endif};
  PathDelim = {$ifdef Darwin}'/'{$else}'\'{$endif};
  faSymLink = $00000040;

const
  // Select the icons used from the Icon bitmap
  // Note: 0 = yellow folder
  //       2 = yellow Win8 like
  IconFolder = 3;  // Blue folder
  IconFile = 1;    // The Midi/Music Xml icon

type

  { TFormOpenDialog }

  TFormOpenDialog = class(TForm)
    BevelFilename: TBevel;
    BevelTreeView: TBevel;
    ButtonOk: TButton;
    EditFileName: TEdit;
    ImageFolders: TImage;
    Timer: TTimer;
    TreeView: TTreeView;
    procedure ButtonOkClick(Sender: TObject);
    procedure OnMouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer);
    procedure OnEdited(Sender: TObject; Node: TTreeNode; var S: string);
    procedure OnActivate(Sender: TObject);
    procedure OnClose(Sender: TObject; var CloseAction: TCloseAction);
    procedure OnCreate(Sender: TObject);
    procedure OnDblClick(Sender: TObject);
    procedure OnResize(Sender: TObject);
    procedure TimerTimer(Sender: TObject);
  private
    { private declarations }
    SelectedFileName: string;
    SetVisible: boolean;
    procedure FindFiles(Dir: string; RootNode: TTreeNode);
    function DirNotEmpty(Dir: string; RootNode: TTreeNode): boolean;
    function AddNode(Node: TTreeNode; Name: string): TTreeNode;
    function AddSubNode(Node: TTreeNode; Name: string;
                        IconNr: integer; Description: string): TTreeNode;
    function ShortPath(Name: string): string;
    procedure SavePosition;
  public
    { public declarations }
  end;

var
  FormOpenDialog: TFormOpenDialog;

implementation

{$R *.lfm}

{ TFormOpenDialog }

//---------------------------------------------------------------------------
//
//     Function:   AddNode
//
//     Purpose:    To add a node to the tree view. Show a folder icon.
//
//     Parameters: Node = the tree view node
//                 Name = text shown (music or download)
//
//     Returns:    void
//
//     Notes:      <none>
//
//---------------------------------------------------------------------------

function TFormOpenDialog.AddNode(Node: TTreeNode; Name: string): TTreeNode;
var
  NewNode: TTreeNode;

  begin
  NewNode:=TreeView.Items.Add(Node,Name);
  NewNode.ImageIndex:=IconFolder;
  NewNode.SelectedIndex:=IconFolder;
  NewNode.Data:=nil;
  AddNode:=NewNode;
  end;

//---------------------------------------------------------------------------
//
//     Function:   AddSubNode
//
//     Purpose:    To add a subnode to the tree view. Show a folder or
//                 file icon.
//
//     Parameters: Node = the tree view node
//                 Name = text shown (music or download)
//                 IconNr = The icon according to the icon file.
//                 (May be different on PC and MAC. MAC is p.t. blue)
//                 See definitions IconFolder and IconFile
//
//     Returns:    void
//
//     Notes:      <none>
//
//---------------------------------------------------------------------------
function TFormOpenDialog.AddSubNode(Node: TTreeNode; Name: string;
  IconNr: integer; Description: string): TTreeNode;

var
  RootNode: TTreeNode;
  s: string;
  PDescription: ^string;

  begin
  RootNode:=Node;
  Node:=TreeView.Items.AddChild(RootNode,Name);
  Node.ImageIndex:=IconNr;
  Node.SelectedIndex:=IconNr;
  New(PDescription);                 //// OBS: Add freeing the memory !!!!
  PDescription^:=Description;
  Node.Data:=PDescription;
  RootNode.Expand(true);
  AddSubNode:=Node;
  end;

//---------------------------------------------------------------------------
//
//     Function:   OnCreate
//
//     Purpose:    To handle the standard windows event OnCreate
//                 If a file Default.txt exists then this file contains
//                 a position and size of this window. After this initialise
//                 the tree view of the directories for Music and Download
//
//     Parameters: Sender = Windows standard
//
//     Returns:    void
//
//     Notes:      <none>
//
//---------------------------------------------------------------------------
var
  log: text;

procedure TFormOpenDialog.OnCreate(Sender: TObject);
var
  ImageList: TImageList;
  Home: string;
  WindowPosition: TextFile;
  L,T,W,H: integer;           // Local variables: Left, Top, Width, Height

  NodeMusicTop: TTreeNode;
  NodeDownloadTop: TTreeNode;

  logname: string;

  begin
  L:=Left;
  T:=Top;
  W:=Width;
  H:=Height;
  SelectedFileName:='';
  EditFileName.Text:='';
  Home:=GetEnvironmentVariable('HOME');
  if FileExists(Home+'/MidiOptions/Default.txt') then
    begin
    AssignFile(WindowPosition,Home+'/MidiOptions/Default.txt');
    reset(WindowPosition);
    if not eof(WindowPosition) then readln(WindowPosition,L);
    if not eof(WindowPosition) then readln(WindowPosition,T);
    if not eof(WindowPosition) then readln(WindowPosition,W);
    if not eof(WindowPosition) then readln(WindowPosition,H);
    CloseFile(WindowPosition);
    end;
  Left:=L;
  Top:=T;
  Width:=W;
  Height:=H;
  ImageList:=TImageList.Create(nil);
  ImageList.Clear;
  ImageList.Add(ImageFolders.Picture.Bitmap,ImageFolders.Picture.Bitmap);
  TreeView.Images:=ImageList;
  TreeView.Items.Clear;
  NodeMusicTop:=AddNode(nil,'Music');
  NodeDownloadTop:=AddNode(nil,'Download');
  logname:=Home+'/find.txt';
  system.assign(log,logname);
  rewrite(log);
  FindFiles(Home+'/Music/',NodeMusicTop);
  FindFiles(Home+'/Downloads/',NodeDownloadTop);
  system.close(log);
  SelectedFileName:='';
  TreeView.FullExpand;
  end;

//---------------------------------------------------------------------------
//
//     Function:   OnDblClick
//
//     Purpose:    To handle the standard windows event OnDblClick
//                 The file is opened directly
//
//     Parameters: Sender = Windows standard
//
//     Returns:    void
//
//     Notes:      <none>
//
//---------------------------------------------------------------------------

procedure TFormOpenDialog.OnDblClick(Sender: TObject);

var
  Node: TTreeNode;
  PDescription: ^string;
  CloseAction: TCloseAction;

  begin
  Errors:=0;
  Node:=TreeView.Selected;
  if Node<>nil then
    begin
    PDescription:=Node.Data;
    if (PDescription<>nil) and (PDescription^<>'') and
       ((LowerCase(ExtractFileExt(PDescription^))='.mid') or
          (LowerCase(ExtractFileExt(PDescription^))='.xml')) then
      begin
      SelectedFileName:=PDescription^;
      EditFileName.Text:=ShortPath(PDescription^);
      ButtonOkClick(Sender);
      end
    else
      begin
      SelectedFileName:='';
      EditFileName.Text:='';
      end;
    end;
  SavePosition;
  end;

//---------------------------------------------------------------------------
//
//     Function:   OnResize
//
//     Purpose:    To handle the standard windows event OnResize
//                 This window can be of any size and the elements
//                 are resized accordingly
//
//     Parameters: Sender = Windows standard
//
//     Returns:    void
//
//     Notes:      <none>
//
//---------------------------------------------------------------------------

procedure TFormOpenDialog.OnResize(Sender: TObject);

    begin
    VertScrollBar.Visible:=false;
{
    ButtonOk.Top:=ClientHeight-(3*EditFileName.Height) div 2;
    ButtonOk.Width:=ClientWidth div 10;
    ButtonOk.Height:=EditFileName.Height;
    ButtonOk.Left:=ClientWidth-ButtonOk.Width-6*GLUE;
}
    EditFileName.Top:=ClientHeight-(3*EditFileName.Height) div 2;
    EditFileName.Left:=2*GLUE;
    EditFileName.Width:=(8*ClientWidth) div 10;
    TreeView.Top:=2*GLUE;
    TreeView.Height:=ClientHeight-2*ButtonOk.Height-3*GLUE;
    TreeView.Left:=2*GLUE;
    TreeView.Width:=ClientWidth-4*GLUE;
    BevelTreeView.Top:=Glue;
    BevelTreeView.Left:=Glue;
    BevelTreeView.Width:=ClientWidth-2*Glue;
    BevelTreeView.Height:=TreeView.Height+2*Glue;

    BevelFileName.Top:=ClientHeight-2*EditFileName.Height+GLUE;
    BevelFileName.Left:=Glue;
    BevelFileName.Width:=ClientWidth-2*Glue;
    BevelFileName.Height:=2*EditFileName.Height-2*GLUE;
    end;


procedure TFormOpenDialog.OnActivate(Sender: TObject);

  begin
  //// P.t. used in the PC-version but...
  end;

//---------------------------------------------------------------------------
//
//     Function:   OnClose
//
//     Purpose:    To handle the standard windows event OnClose
//                 The position and size of the window is saved
//
//     Parameters: Sender = Windows standard
//
//     Returns:    void
//
//     Notes:      <none>
//
//---------------------------------------------------------------------------

procedure TFormOpenDialog.OnClose(Sender: TObject; var CloseAction: TCloseAction);

  begin
  SavePosition;
  end;

//---------------------------------------------------------------------------
//
//     Function:   SavePosition
//
//     Purpose:    The save the position and size of the window
//                 in order to show it at same place next time
//
//     Parameters: <none>
//
//     Returns:    void
//
//     Notes:      <none>
//
//---------------------------------------------------------------------------

procedure TFormOpenDialog.SavePosition;
  var
    WindowPosition: TextFile;
    Home: string;

    begin
    Home:=GetEnvironmentVariable('HOME');
    if not DirectoryExists(Home+'/MidiOptions') then
      MkDir(Home+'/MidiOptions');
    AssignFile(WindowPosition,Home+'/MidiOptions/Default.txt');
    rewrite(WindowPosition);
    writeln(WindowPosition,Left);
    writeln(WindowPosition,Top);
    writeln(WindowPosition,Width);
    writeln(WindowPosition,Height);
    CloseFile(WindowPosition);
    end;

//---------------------------------------------------------------------------
//
//     Function:   ButtonOkClick
//
//     Purpose:    To handle the standard windows event ButtonOkClick
//                 By mouse click the the is selected and show in the edit
//                 field below the file tree view. The file is opened directly
//                 The system state is changed to be ready to start playing
//
//     Parameters: Sender = Windows standard
//
//     Returns:    void
//
//     Notes:      <none>
//
//---------------------------------------------------------------------------

procedure TFormOpenDialog.ButtonOkClick(Sender: TObject);

    begin
    Errors:=0;
    if FileExistsUtf8(SelectedFileName) then
      begin
      FileName:=SelectedFileName;
      FormMessage.Create;
      SystemState:=MidiStarting;
      end
    else
      begin
      SystemState:=MidiNoFile;
      end;
    Visible:=false;
    SavePosition;
    end;

//---------------------------------------------------------------------------
//
//     Function:   OnMouseMove
//
//     Purpose:    To handle the standard windows event OnMouseMove
//                 NOT USED. PREPARE DRAG AND DROP
//
//     Parameters: Sender = Windows standard
//                 shift  = state of CTRL and ALT and SHIFT buttons
//                 X      = x-position of mouse pressed
//                 Y      = y-position of mouse pressed
//
//     Returns:    void
//
//     Notes:      <none>
//
//---------------------------------------------------------------------------

procedure TFormOpenDialog.OnMouseMove(Sender: TObject; Shift: TShiftState; X,
  Y: Integer);

var DragMode: TDragMode;
    DragKind: TDragKind;
    FileName: string;
    ////FileObject: TFileObject;

  begin
  ////DataObject:=GetFileListDataObject(Directory, SelFileList);
  ////DoDragDrop(DataObject, Self, DROPEFFECT_COPY, Effect);
  ////CopyFilesToClipboard();

  TreeView.DragMode:=dmManual;
  TreeView.DragCursor:=crDrag;
  BeginDrag(true);
  TreeView.DragCursor:=crDrag;
  TreeView.DragKind:=dkDrag;
  ////TreeView.DragDrop(FileName,1,1);
  end;

//---------------------------------------------------------------------------
//
//     Function:   OnEdited
//
//     Purpose:    To handle the standard windows event OnEdited
//                 A name in the tree view may be edited. Change
//                 the file name to the new one
//
//     Parameters: Sender = Windows standard
//
//     Returns:    void
//
//     Notes:      <none>
//
//---------------------------------------------------------------------------

procedure TFormOpenDialog.OnEdited(Sender: TObject; Node: TTreeNode;var S: string);

  var
    OldFileName: string;         //// description !!!!
    NewFileName: string;
    OldFilePath: string;
    NewFilePath: string;
    PDescription: ^string;
    p: integer; // Position of delimiter '/'
    i: integer; // Loop variable  for positions

    begin
    PDescription:=Node.Data;
    OldFilePath:=PDescription^;
    OldFileName:=ExtractFileName(OldFilePath);
    // Find the path part
    p:=0;
    for i:=1 to Length(OldFilePath) do if OldFilePath[i]='/' then p:=i;
    NewFilePath:=copy(OldFilePath,1,p);
    NewFileName:=S;
    NewFilePath:=NewFilePath+NewFileName;
    RenameFile(OldFilePath,NewFilePath);
    PDescription^:=NewFilePath;
    SelectedFileName:=PDescription^;
    EditFileName.Text:=ShortPath(PDescription^);
    end;

//---------------------------------------------------------------------------
//
//     Function:   TimerTimer
//
//     Purpose:    To handle the standard windows event TimerTimer
//                 When a tree item is selected, there seem to be an
//                 updating problem - sometimes a wrong name appears
//                 (the previous selected item). Therefore the timer
//                 assures that the name will be right at next timer
//                 click.
//
//     Parameters: Sender = Windows standard
//
//     Returns:    void
//
//     Notes:      <none>
//
//---------------------------------------------------------------------------

procedure TFormOpenDialog.TimerTimer(Sender: TObject);
var
  Node: TTreeNode;
  PDescription: ^string;

  begin
  Node:=TreeView.Selected;
  if Node<>nil then
    begin
    PDescription:=Node.Data;
    if (PDescription<>nil) and (PDescription^<>'') and
       ((LowerCase(ExtractFileExt(PDescription^))='.mid') or
          (LowerCase(ExtractFileExt(PDescription^))='.xml')) then
      begin
      SelectedFileName:=PDescription^;
      EditFileName.Text:=ShortPath(PDescription^);
      end;
    end;
  end;

//---------------------------------------------------------------------------
//
//     Function:   FindFiles
//
//     Purpose:    To traverse the whole file structure and add the
//                 elements to the tree view.
//
//     Parameters: Dir = current directory
//                 RootNode = current node in the tree view
//
//     Returns:    void
//
//     Notes:      <none>
//
//---------------------------------------------------------------------------

procedure TFormOpenDialog.FindFiles(Dir: string; RootNode: TTreeNode);

var
  Info: TSearchRec;
  FileName: string;
  Node: TTreeNode;

  MusicXmlFile: textfile;
  s: string;

  i: integer;
  begin
  writeln(log,'DefaultSystemCodePage:',DefaultSystemCodePage,' DefaultFileSystemCodePage=',DefaultFileSystemCodePage,' DefaultRTLFileSystemCodePage=',DefaultRTLFileSystemCodePage);
  writeln(log,'LC_ALL=',GetEnvironmentVariable('LC_ALL'),' LC_MESSAGES=',GetEnvironmentVariable('LC_MESSAGES'),' LANG=',GetEnvironmentVariable('LANG'));
  writeln(log,'dir code page: ',stringcodepage(dir));
  // Find files in the current directory
  if FindFirstUTF8(Dir+FileName,faAnyFile and not faDirectory,Info) { *Converted from FindFirst* } = 0 then
  try
    repeat
      // Work with found file info...
    until FindNextUTF8(Info) { *Converted from FindNext* }<>0;
  finally
    FindCloseUTF8(Info); { *Converted from FindClose* }
  end;
  // Find subdirectories and recurse
  if FindFirstUTF8(Dir+SAnyMask,faDirectory,Info) { *Converted from FindFirst* }=0 then
    try
      repeat
      if (Info.Attr and faDirectory<>0) and
         (Info.Name<>'.') and (Info.Name<>'..') and
         (Info.Attr and faSymLink=0) then
        begin // Dir - add name to directory path
        if DirNotEmpty(Dir+Info.Name+PathDelim,Node) then
        Node:=AddSubNode(RootNode,Info.Name,IconFolder,Dir+Info.Name);
        FindFiles(Dir+Info.Name+PathDelim,Node);
        end
      else
        begin // File - only midi and music xml.
        if ((Info.Name<>'.') and (Info.Name<>'..')) and
           (LowerCase(ExtractFileExt(Info.Name))='.mid') or
           (LowerCase(ExtractFileExt(Info.Name))='.xml') then
          begin
          if LowerCase(ExtractFileExt(Info.Name))='.xml' then
            begin
            {$i-}
////            ChDir(Dir);
////            AssignFile(MusicXmlFile,Info.Name);
            writeln(log,'file = ',Dir+Info.Name,' -- code page: ',stringcodepage(Dir+Info.Name));
            AssignFile(MusicXmlFile,Dir+Info.Name);
            if IORESULT=0 then
            reset(MusicXmlFile);
            {$i+}
            if IORESULT=0 then
              begin
              // A music xml file has the word MusicXML in second line
              if not eof(MusicXmlFile) then readln(MusicXmlFile,s);
              if not eof(MusicXmlFile) then readln(MusicXmlFile,s);
              if Pos('MusicXML',s)>0 then
              Node:=AddSubNode(RootNode, Info.Name,IconFile,Dir+Info.Name);
              CloseFile(MusicXmlFile);
              end;
            end
          else
            Node:=AddSubNode(RootNode, Info.Name,1,Dir+Info.Name);
          end;
        end;
      until FindNextUTF8(Info) { *Converted from FindNext* }<>0;
    finally
    FindCloseUTF8(Info); { *Converted from FindClose* }
    end;
  end;

//---------------------------------------------------------------------------
//
//     Function:   DirNotEmpty
//
//     Purpose:    To traverse the whole file structure and test
//                 if directory is empty (i.e. contains no Midi nor Music Xml)
//
//     Parameters: Dir = current directory
//                 RootNode = current node in the tree view
//
//     Returns:    void
//
//     Notes:      <none>
//
//---------------------------------------------------------------------------

function TFormOpenDialog.DirNotEmpty(Dir: string; RootNode: TTreeNode): boolean;

var
  Info: TSearchRec;
  FileName: string;
  Node: TTreeNode;
  i: integer;
  NotEmpty: boolean;
  MusicXmlFile: textfile;
  s: string;

  begin
  // Find files in the current directory...
  if FindFirstUTF8(Dir+FileName,faAnyFile and not faDirectory,Info) { *Converted from FindFirst* } = 0 then
  try
    repeat
      // Work with found file info...
    until FindNextUTF8(Info) { *Converted from FindNext* }<>0;
  finally
    FindCloseUTF8(Info); { *Converted from FindClose* }
  end;
  // Find subdirectories and recurse
  if FindFirstUTF8(Dir+SAnyMask,faDirectory,Info) { *Converted from FindFirst* }=0 then
    try
      NotEmpty:=false;
      repeat
      if (Info.Attr and faDirectory<>0) and
         (Info.Name<>'.') and (Info.Name<>'..') and
         (Info.Attr and faSymLink=0) then
        begin // Dir
        NotEmpty:=NotEmpty or DirNotEmpty(Dir+Info.Name+PathDelim,Node);
        end
      else
        begin // File
{
        FileName:=Dir+'xyzæøå.test';
        AssignFile(MusicXmlFile,FileName);
        rewrite(MusicXmlFile);
        CloseFile(MusicXmlFile);
}
        if ((Info.Name<>'.') and (Info.Name<>'..')) and
           (LowerCase(ExtractFileExt(Info.Name))='.mid') or
           (LowerCase(ExtractFileExt(Info.Name))='.xml') then
          begin
          if LowerCase(ExtractFileExt(Info.Name))='.xml' then
            begin
            {$i-}
            AssignFile(MusicXmlFile,PChar(Dir+Info.Name));
            if IORESULT=0 then
            reset(MusicXmlFile);
            {$i+}
            if IORESULT=0 then
              begin
              if not eof(MusicXmlFile) then readln(MusicXmlFile,s);
              if not eof(MusicXmlFile) then readln(MusicXmlFile,s);
              if Pos('MusicXML',s)>0 then NotEmpty:=true;
              CloseFile(MusicXmlFile);
              end;
            end
          else
            NotEmpty:=true;
          end;
        end;
      until FindNextUTF8(Info) { *Converted from FindNext* }<>0;
    finally
    FindCloseUTF8(Info); { *Converted from FindClose* }
    end;
    DirNotEmpty:=NotEmpty;
  end;

//---------------------------------------------------------------------------
//
//     Function:   ShortPath
//
//     Purpose:    In order to avoid the very long and strange
//                names in the Apple's sandboxing environment
//                 only show 'Music' and 'Download' ∫and the rest
//
//     Parameters: Name = the real file name with path
//
//     Returns:    A more readable name.
//
//     Notes:      <none>
//
//---------------------------------------------------------------------------

function TFormOpenDialog.ShortPath(Name: string): string;

var
  Home: string;
  p: integer;

  begin
  Home:=GetEnvironmentVariable('HOME');
  p:=Pos(Home,Name);
  // Pass the HOME and '/'
  if p=1 then ShortPath:=copy(Name,Length(Home)+2,Length(Name)-Length(Home));
  end;


end.








